#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "mpi.h"
#include "Comm.h"

#ifndef _LALSTDLIB_H
#include "LALStdlib.h"
#ifndef _LALSTDLIB_H
#define _LALSTDLIB_H
#endif
#endif

static CHAR *UTILS="$Id: Preliminary version of communications library$";

static CHAR* messagetypes[]=MESSAGETYPELIST;

/* The different allowed message tags (for robustness) */
#define MESSTYPE 1
#define TEMPTYPE 2
#define DATATYPE 3
#define EVENTYPE 4
#define REQUESTTYPE 5

void ExchangeMessage(Status* status,MessageExchange* params) {
    INT4 returnvalue;
    
        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
        /* if sending a message to the master */
    if (params->send){
            /* make sure that the message is one that is recognized */
        ASSERT(params->message>=0 && params->message<NUMBERMESSAGES,status,COMM_UNRECMESG,
                      COMM_MSGUNRECMESG);
        returnvalue=MPI_Send(&(params->message),1,MPI_INT,0,MESSTYPE,MPI_COMM_WORLD);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
    else {
            /* we are getting a message type from a slave */
        MPI_Status mpistatus;
        returnvalue=MPI_Recv(&(params->message),1,MPI_INT,MPI_ANY_SOURCE,MESSTYPE,MPI_COMM_WORLD,&mpistatus);
        returnvalue=(returnvalue==MPI_SUCCESS) && (mpistatus.MPI_ERROR==0);
        ASSERT(returnvalue,status,COMM_MPIERROR,COMM_MSGMPIERROR);
            /* make sure that the message is one that is recognized */
        ASSERT(params->message>=0 && params->message<NUMBERMESSAGES,status,
                      COMM_UNRECMESG,COMM_MSGUNRECMESG);
        params->source=mpistatus.MPI_SOURCE;
        params->type=messagetypes[params->message];
    }
    
        /* normal exit */
    RETURN (status);
}

/* NOTE: this routine assumes same architecture in master and slave */
void ExchangeTemplates(Status* status,TemplateExchange* params) 
{
    INT4 number,sizeinbytes,returnvalue;
    
        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
        /* are we sending templates? */
    if (params->send) {
            /* number of templates */
        number=params->number;
        ASSERT(number>=0,status,COMM_NEGATIVE,COMM_MSGNEGATIVE);
            /* size in bytes of object being sent */
        sizeinbytes=number*sizeof(tempstruct);
            /* if positive number of objects, make sure they are there! */
        if (number>0) {
            ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
	}
            /* and send them */
        returnvalue=MPI_Send(params->list,sizeinbytes,MPI_BYTE,params->destination,
                             TEMPTYPE,MPI_COMM_WORLD);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
    else {
        MPI_Status mpistatus;
            /* get details of pending message */
        returnvalue=MPI_Probe(0,MPI_ANY_TAG,MPI_COMM_WORLD,&mpistatus);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
            /* array size to be received */
        returnvalue=MPI_Get_count(&mpistatus,MPI_BYTE,&sizeinbytes);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
        
            /* number of templates */
        number=sizeinbytes/sizeof(tempstruct);
        params->number=number;
        ASSERT(number>=0,status,COMM_NEGATIVE,COMM_MSGNEGATIVE);
            /* allocate storage array and check that it exists */
        params->list=(tempstruct *)LALMalloc(sizeinbytes);
        if (number>0) {
            ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
        }
            /* then get the template list */
        returnvalue=MPI_Recv(params->list,sizeinbytes,MPI_BYTE,0,TEMPTYPE,MPI_COMM_WORLD,&mpistatus);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
        
    }
        /* normal exit */
    RETURN (status);
}

/* NOTE: this routine assumes same architecture in master and slave */
void ExchangeData(Status* status,DataExchange* params) 
{
    INT4 number,sizeinbytes,returnvalue;
    
        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
    
        /* are we sending data? */
    if (params->send) {
            /* number of blocks of data */
        number=params->number;
        ASSERT(number>=0,status,COMM_NEGATIVE,COMM_MSGNEGATIVE);
            /* size in bytes of object being sent */
        sizeinbytes=number*sizeof(DataPacket);
            /* if positive number of objects, make sure they are there! */
        if (number>0) {
            ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
        }
            /* and send them */
        returnvalue=MPI_Send(params->list,sizeinbytes,MPI_BYTE,params->destination,DATATYPE,MPI_COMM_WORLD);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
    else {
        MPI_Status mpistatus;
            /* get details of pending message */
        returnvalue=MPI_Probe(0,DATATYPE,MPI_COMM_WORLD,&mpistatus);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
            /* array size to be received */
        returnvalue=MPI_Get_count(&mpistatus,MPI_BYTE,&sizeinbytes);
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
            /* number of templates */
        number=sizeinbytes/sizeof(DataPacket);
        params->number=number;
            /* allocated storage array and gets template bank */
        params->list=(DataPacket *)LALMalloc(sizeinbytes);
        if (number>0) {
            ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
        }
            /* and send them */
        returnvalue=MPI_Recv(params->list,sizeinbytes,MPI_BYTE,0,DATATYPE,MPI_COMM_WORLD,&mpistatus);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
            /* normal exit */
    RETURN (status);
}

/* NOTE: this routine assumes same architecture in master and slave */
void ExchangeEvents(Status* status,EventExchange* params) 
{
    INT4 number,sizeinbytes,returnvalue;

        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
        /* are we sending data? */
    if (params->send) {
            /* number of events */
        number=params->number;
        ASSERT(number>=0,status,COMM_NEGATIVE,COMM_MSGNEGATIVE);
            /* size in bytes of object being sent */
        sizeinbytes=number*sizeof(EventPacket);
            /* if positive number of objects, make sure they are there! */
        if (number>0) {
            ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
        }
            /* and send them */
        returnvalue=MPI_Send(params->list,sizeinbytes,MPI_BYTE,0,EVENTYPE,MPI_COMM_WORLD);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
    else {
            MPI_Status mpistatus;
                /* get details of pending message */
            returnvalue=MPI_Probe(params->source,EVENTYPE,MPI_COMM_WORLD,&mpistatus);
            ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
                /* array size to be received */
            returnvalue=MPI_Get_count(&mpistatus,MPI_BYTE,&sizeinbytes);
            ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
                 /* number of events */
            number=sizeinbytes/sizeof(EventPacket);
            params->number=number;
                /* allocated storage array and gets template bank */
            params->list=(EventPacket *)LALMalloc(sizeinbytes);
            if (number>0) {
                ASSERT(params->list!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
            }
            returnvalue=MPI_Recv(params->list,sizeinbytes,MPI_BYTE,params->source,
                                 EVENTYPE,MPI_COMM_WORLD,&mpistatus);
                /* check that no MPI errors occured */
            ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
        /* normal exit */
    RETURN (status);
}
/* NOTE: this routine assumes same architecture in master and slave */
void ExchangeRequest(Status* status,RequestExchange* params) 
{
    INT4 returnvalue,sizeinbytes=sizeof(TemplateRequest);
        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);

        /* are we sending the request? */
    if (params->send) {
            /* size in bytes of object being sent */
        returnvalue=MPI_Send(&(params->list),sizeinbytes,MPI_BYTE,0,REQUESTTYPE,MPI_COMM_WORLD);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
    else {
        MPI_Status mpistatus;
            /* get template bank request */
        returnvalue=MPI_Recv(&(params->list),sizeinbytes,MPI_BYTE,params->source,REQUESTTYPE,
                             MPI_COMM_WORLD,&mpistatus);
            /* check that no MPI errors occured */
        ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    }
            /* normal exit */
    RETURN (status);
}

/* This routine exports environment variable params->name to all nodes */
void ExportEnvironment(Status *status,ExportParams* params)
{
    CHAR *environment,*var=NULL;
    INT4 length,returnvalue;
    
        /* initialize status */
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
    
        /* if this is the master, get enviroment variable */
    if (params->identity.myid==0) {
        ASSERT(params->name!=NULL,status,COMM_NULLENV,COMM_MSGNULLENV);
        var=getenv(params->name);
        ASSERT(var!=NULL,status,COMM_NOENV,COMM_MSGNOENV);
            /* calculate the length of the string to pass */
        length=strlen(params->name)+strlen(var)+2;
    }
    
        /* broadcast the length of the string to all nodes */
    returnvalue=MPI_Bcast(&length,1,MPI_INT,0,MPI_COMM_WORLD);
    ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    
        /* allocate storage for the string on all nodes */
    environment=(CHAR *)LALMalloc(length);
    ASSERT(environment!=NULL,status,COMM_NULLBUFF,COMM_MSGNULLBUFF);
    
        /*  copy environment variable into string onto master */
    if (params->identity.myid==0)
        sprintf(environment,"%s=%s",params->name,var);
    
        /* then broadcast it to the other processes */
    returnvalue=MPI_Bcast(environment,length,MPI_CHAR,0,MPI_COMM_WORLD);
    ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    
        /* and put it into the local environment */
    if (params->identity.myid!=0) {
        returnvalue=putenv(environment);
        ASSERT(returnvalue==0,status,COMM_PUTENV,COMM_MSGPUTENV);
            /* printf("Node %d just set %s\n",params->identity.myid,environment); */
    }
        /* normal exit */
    RETURN (status);
}

void DebugMPI(Status* status,DebugMPIParams* params) 
{
    Status estat;
	ExportParams expar;
    CHAR commandstring[256];
    pid_t process_id=getpid();
    INT4 returnvalue;
    
    INITSTATUS(status,UTILS);
        /* check that parameter block is not null */
    ASSERT(params!=NULL,status,COMM_NULLPARA,COMM_MSGNULLPARA);
    
        /* export environment variable to the slaves */
    expar.identity=params->identity;
    expar.name="DISPLAY";
    ExportEnvironment(&estat,&expar);
    ASSERT(estat.statusCode==0,status,COMM_EXPORTFAILED,COMM_MSGEXPORTFAILED);
    
    expar.name="LD_LIBRARY_PATH";
    ExportEnvironment(&estat,&expar);
    ASSERT(estat.statusCode==0,status,COMM_EXPORTFAILED,COMM_MSGEXPORTFAILED);
    
    
        /* then start the debugger if desired */
    if (params->startup) {
        ASSERT(params->debugger!=NULL && params->programname!=NULL,status,
                      COMM_NULLSTRING,COMM_MSGNULLSTRING);
        sprintf(commandstring,"%s %s %d &",params->debugger,params->programname,(INT4)process_id);
        returnvalue=system(commandstring);
        ASSERT(returnvalue==0,status,COMM_SUBPROC,COMM_MSGSUBPROC);
        
    }
    
    	/* time to start up on each node */
    if (params->delaytime>0)
        sleep(params->delaytime);
    
        /* synchronize everyone! */
    returnvalue=MPI_Barrier(MPI_COMM_WORLD);
    ASSERT(returnvalue==MPI_SUCCESS,status,COMM_MPIERROR,COMM_MSGMPIERROR);
    
        /* normal exit */
    RETURN (status);
}
