#ifndef _STDIO_H
#include <stdio.h>
#ifndef _STDIO_H
#define _STDIO_H
#endif
#endif

#ifndef _MATH_H
#include <math.h>
#ifndef _MATH_H
#define _MATH_H
#endif
#endif

#ifndef _LALSTDLIB_H
#include "LALStdlib.h"
#ifndef _LALSTDLIB_H
#define _LALSTDLIB_H
#endif
#endif

#ifndef _AVFACTORIES_H
#include "AVFactories.h"
#ifndef _AVFACTORIES_H
#define _AVFACTORIES_H
#endif
#endif

#ifndef _DATABUFFER_H
#include "DataBuffer.h"
#ifndef _DATABUFFER_H
#define _DATABUFFER_H
#endif
#endif

int debuglevel = 1;

int
main ()
{
  const INT4 numPoints = 65536;
  const INT4 numSpec   = 8;
  const INT4 numSegs   = 10;

  Status status;

  INT4 i;
  INT4 seg;
  char fname[256];
  FILE *fp;

  DataBuffer    *buffer = NULL;
  DataBufferPar  bufferPar;
  GetDataOut     dataout;

  INT2TimeSeries          data;
  REAL4FrequencySeries    spec;
  COMPLEX8FrequencySeries resp;

  data.data = NULL;
  spec.data = NULL;
  resp.data = NULL;
  I2CreateVector (&status, &data.data, numPoints);
  CreateVector (&status, &spec.data, numPoints/2 + 1);
  CCreateVector (&status, &resp.data, numPoints);
  dataout.data = &data;
  dataout.spec = &spec;
  dataout.resp = &resp;

  bufferPar.numSpec    = numSpec;
  bufferPar.numPoints  = numPoints;
  bufferPar.windowType = Welch;
  bufferPar.plan       = NULL;
  EstimateFwdRealFFTPlan (&status, &bufferPar.plan, numPoints);

  CreateDataBuffer (&status, &buffer, &bufferPar);

  for (seg = 0; seg < numSegs; ++seg)
  {
    fprintf (stderr, "Segment: %03d\n", seg);

    GetData (&status, &dataout, 3*numPoints/4, buffer);
    if (status.statusCode)
      goto exit;

    /* print data vector */
    if (0)
    {
      sprintf (fname, "Segment.%03d", seg);
      fp = fopen (fname, "w");

      for (i = 0; i < dataout.data->data->length; ++i)
        fprintf (fp, "%u\t%d\n", i, dataout.data->data->data[i]);

      fclose (fp);
    }

    if (0)
    {
      sprintf (fname, "Spectrum.%03d", seg);
      fp = fopen (fname, "w");

      for (i = 0; i < dataout.spec->data->length; ++i)
        fprintf (fp, "%u\t%e\n", i, dataout.spec->data->data[i]);

      fclose (fp);
    }

    /* print calibration info */
    if (1)
    {
      sprintf (fname, "Response.%03d", seg);
      fp = fopen (fname, "w");

      for (i = 0; i < dataout.resp->data->length; ++i)
      {
        REAL8 re  = dataout.resp->data->data[i].re;
        REAL8 im  = dataout.resp->data->data[i].im;
        REAL8 mod = sqrt (re*re + im*im);
        REAL8 arg = atan2 (im, re);
        fprintf (fp, "%u\t%e\t%e\n", i, mod, arg);
      }

      fclose (fp);
    }

  }

exit:

  DestroyRealFFTPlan (&status, &bufferPar.plan);
  DestroyDataBuffer (&status, &buffer);
  I2DestroyVector (&status, &data.data);
  DestroyVector (&status, &spec.data);
  CDestroyVector (&status, &resp.data);

  LALCheckMemoryLeaks ();
  return 0;
}
